/*
 *  lhii.c
 *
 *  Copyright (c) 1999 Harald Bauer  (bauer-h@web.de)
 *
 */

/* 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/****************************************************************************

   This file implements a SANE backend for handheld scanners that use LHII.

   (feedback to:  bauer-h@web.de)

 ***************************************************************************/


#define LHII_MAJOR 0
#define LHII_MINOR 20

#include <sys/ioctl.h>
#include <time.h>     
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include <math.h>

#include <sane/sane.h>
#include <sane/sanei.h>
#include <sane/sanei_config.h>
#include <sane/saneopts.h>

#define BACKEND_NAME lhii
#include <sane/sanei_backend.h>

#define LHII_CONFIG_FILE "lhii.conf"

#ifndef PATH_MAX
# define PATH_MAX	1024
#endif

#define LHII_CAP_2_SANE_CAP(x) ((((x & 0x0002)<<1) | ((x & 0x0004)>>1)) | \
				(x & 0xfff9))

enum lhii_Option
{
  OPT_NUM_OPTS = 0,
  OPT_MODE_GROUP,
  OPT_MONO,
  OPT_GREY,
  OPT_TRUE,
  OPT_XDPI,
  OPT_YDPI,
  OPT_ENHANCEMENT_GROUP,
  OPT_BRIGHT,
  OPT_CONTRAST,
  OPT_HUE,
  OPT_SAT,
  OPT_GEOMETRY_GROUP,
  OPT_LENGTH,
  OPT_ADDITIONAL_GROUP,
  NUM_REG_OPT
};

#include <scanner.h>

struct lhii_device_list_element
{
  struct lhii_device_list_element *next_device;
  SANE_Device sane_device;
};

struct lhii_devstruct
{
  int fd;
  int opt_count;
  SANE_String dev_name; /* name of the current selected device */
  SANE_Int num_lines;   /* number of lines to be scanned */
  SANE_Int scan_length; /* scanning length in mm */
  SANE_Int bytes2read;  /* number of bytes to be scanned */
  struct modeinfo_struct mode;
  SANE_Option_Descriptor *Option_Descriptor;
  struct lhii_device_list_element *lhii_device_list;
  const SANE_Device **sane_device_list;
};

static SANE_Handle lhii_sane_handle;


/********************************************************************/
/********************************************************************/
/*** Wrapper-functions for the ioctl-commands of LHII ***************/
/********************************************************************/
/********************************************************************/


/********************************************************************/
/* open_scanner_device                                              */
/* open the device special file that belongs to the LHII-scanner    */
/********************************************************************/
static SANE_Status
open_scanner_device(const char* dev_name,int* fd)
{
  if ((*fd = open(dev_name,O_RDONLY)) == -1)
    {
      return SANE_STATUS_IO_ERROR;
    }
  return SANE_STATUS_GOOD; 
}

/********************************************************************/
/* close_scanner_device                                             */
/* close the device special file that belongs to the LHII-scanner   */
/********************************************************************/
static SANE_Status
close_scanner_device(int fd)
{
  if (close(fd) == -1)
    {
      return SANE_STATUS_IO_ERROR;
    }
  return SANE_STATUS_GOOD; 
}

/********************************************************************/
/* get_scanner_type                                                 */
/* getting the scanner-type given in the scanner_type struct        */
/********************************************************************/
static SANE_Status
get_scanner_type(int fd,struct scanner_type *type)
{       
  if (ioctl(fd,HSIOCGSCT,type) == -1)
    {
      return SANE_STATUS_IO_ERROR;
    }
  return SANE_STATUS_GOOD; 
}

/********************************************************************/
/* get_scanner_mode                                                 */
/* getting the scan-mode given in the modeinfo_struct struct        */
/********************************************************************/
static SANE_Status
get_scanner_mode(int fd,struct modeinfo_struct *mode)
{       
   if (ioctl(fd,HSIOCGMOD,mode) == -1)
   {
      return SANE_STATUS_IO_ERROR;
   }
   return SANE_STATUS_GOOD; 
}

/********************************************************************/
/* set_scanner_mode                                                 */
/* setting the scan-mode given in the modeinfo_struct struct        */
/********************************************************************/
static SANE_Status
set_scanner_mode(int fd,struct modeinfo_struct *mode)
{       
   if (ioctl(fd,HSIOCSMOD,mode) == -1)
   {
      return SANE_STATUS_IO_ERROR;
   }
   return SANE_STATUS_GOOD; 
}

/********************************************************************/
/* get_scanner_options                                              */
/* getting the additional options given through an array of         */
/* scanner_option struct's                                          */
/********************************************************************/
static SANE_Status
get_scanner_options(int fd,struct scanner_option *options)
{  
   if (ioctl(fd,HSIOCGOPT,options) == -1)
   {
     return SANE_STATUS_IO_ERROR;
   }
   return SANE_STATUS_GOOD; 
}

/********************************************************************/
/* get_scanner_capabilities                                         */
/* getting the scanner-capabilities given in                        */ 
/* the scanner_capabilities struct                                  */
/********************************************************************/
static SANE_Status
get_scanner_capabilities(int fd,struct scanner_capabilities *capabilities)
{       
   if (ioctl(fd,HSIOCGSCC,capabilities) == -1)
   {
      return SANE_STATUS_IO_ERROR;
   }
   return SANE_STATUS_GOOD; 
}


/********************************************************************/
/********************************************************************/
/*** Utility Functions **********************************************/
/********************************************************************/
/********************************************************************/


/********************************************************************/
/* destruct_Option_Descriptor_list                                  */
/* free the memmory allocated by an Option_Descriptor-list          */
/********************************************************************/
void
destruct_Option_Descriptor_list (SANE_Handle h)
{
  int i = 0;
  struct lhii_devstruct *handle = h;
  
  DBG_INIT();
  DBG(1, "destruct_Option_Descriptor_list: begin\n");
  for (i=0;i<handle->opt_count;i++)
    {
      if (handle->Option_Descriptor[i].constraint.range)
	free ((void*)handle->Option_Descriptor[i].constraint.range);
      if ((i >= NUM_REG_OPT) && handle->Option_Descriptor[i].name)
	free ((void*)handle->Option_Descriptor[i].name);
      if ((i >= NUM_REG_OPT) && handle->Option_Descriptor[i].title)
	free ((void*)handle->Option_Descriptor[i].title);
    }
  if (handle->Option_Descriptor)
    free (handle->Option_Descriptor);
  DBG(1, "destruct_Option_Descriptor_list: end\n");
}

/********************************************************************/
/* construct_Option_Descriptor_list                                 */
/* create an Option_Descriptor-list with the data given by the      */
/* scanner_capabilities and the scanner_option of the LHII-scanner  */
/********************************************************************/
static SANE_Status
construct_Option_Descriptor_list (SANE_Handle h)
{
  int i;
  struct lhii_devstruct *handle = h;
  static SANE_Status status = SANE_STATUS_IO_ERROR;
  struct scanner_capabilities capabilities;
  struct scanner_option *options;
  SANE_Range *range;

  status = get_scanner_capabilities(handle->fd,&capabilities);
  if (status == SANE_STATUS_IO_ERROR)
    return status;

  handle->opt_count = NUM_REG_OPT + capabilities.av_options;

  if (!(handle->Option_Descriptor = 
	malloc (handle->opt_count*sizeof(SANE_Option_Descriptor))))
    {
      destruct_Option_Descriptor_list (handle);
      return SANE_STATUS_NO_MEM;
    }
  memset(handle->Option_Descriptor, 0, 
	 handle->opt_count*sizeof(SANE_Option_Descriptor));

  /* regular options */
  handle->Option_Descriptor[OPT_NUM_OPTS].name = SANE_NAME_NUM_OPTIONS;
  handle->Option_Descriptor[OPT_NUM_OPTS].title = SANE_TITLE_NUM_OPTIONS;
  handle->Option_Descriptor[OPT_NUM_OPTS].desc = SANE_DESC_NUM_OPTIONS;
  handle->Option_Descriptor[OPT_NUM_OPTS].type = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_NUM_OPTS].unit = SANE_UNIT_NONE;
  handle->Option_Descriptor[OPT_NUM_OPTS].size = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_NUM_OPTS].cap  = SANE_CAP_SOFT_DETECT;
  handle->Option_Descriptor[OPT_NUM_OPTS].constraint_type = 
    SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_MODE_GROUP].name  = "";
  handle->Option_Descriptor[OPT_MODE_GROUP].title = "Scan Mode";
  handle->Option_Descriptor[OPT_MODE_GROUP].desc  = "";
  handle->Option_Descriptor[OPT_MODE_GROUP].type  = SANE_TYPE_GROUP;
  handle->Option_Descriptor[OPT_MODE_GROUP].cap   = 0;
  handle->Option_Descriptor[OPT_MODE_GROUP].constraint_type = 
    SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_MONO].name = "mono";
  handle->Option_Descriptor[OPT_MONO].title = "B/W scan";
  handle->Option_Descriptor[OPT_MONO].desc = 
    "scanning with a depth of 1 bit per pixel in B/W-mode";
  handle->Option_Descriptor[OPT_MONO].type = SANE_TYPE_BOOL;
  handle->Option_Descriptor[OPT_MONO].unit = SANE_UNIT_NONE;
  handle->Option_Descriptor[OPT_MONO].size = sizeof(SANE_Bool);
  handle->Option_Descriptor[OPT_MONO].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_scm_mono);
  handle->Option_Descriptor[OPT_MONO].constraint_type = SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_GREY].name = "grey";
  handle->Option_Descriptor[OPT_GREY].title = "grey scan";
  handle->Option_Descriptor[OPT_GREY].desc = 
    "scanning with a depth of 8 bit per pixel in grey-scale-mode";
  handle->Option_Descriptor[OPT_GREY].type = SANE_TYPE_BOOL;
  handle->Option_Descriptor[OPT_GREY].unit = SANE_UNIT_NONE;
  handle->Option_Descriptor[OPT_GREY].size = sizeof(SANE_Bool);
  handle->Option_Descriptor[OPT_GREY].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_scm_grey);
  handle->Option_Descriptor[OPT_GREY].constraint_type = SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_TRUE].name = "true";
  handle->Option_Descriptor[OPT_TRUE].title = "true scan";
  handle->Option_Descriptor[OPT_TRUE].desc = 
    "scanning with a depth of 24 bit per pixel in true-color-mode";
  handle->Option_Descriptor[OPT_TRUE].type = SANE_TYPE_BOOL;
  handle->Option_Descriptor[OPT_TRUE].unit = SANE_UNIT_NONE;
  handle->Option_Descriptor[OPT_TRUE].size = sizeof(SANE_Bool);
  handle->Option_Descriptor[OPT_TRUE].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_scm_true);
  handle->Option_Descriptor[OPT_TRUE].constraint_type = SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_XDPI].name  = SANE_NAME_SCAN_X_RESOLUTION;
  handle->Option_Descriptor[OPT_XDPI].title = SANE_TITLE_SCAN_X_RESOLUTION;
  handle->Option_Descriptor[OPT_XDPI].desc  = SANE_DESC_SCAN_X_RESOLUTION;
  handle->Option_Descriptor[OPT_XDPI].type  = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_XDPI].unit  = SANE_UNIT_DPI;
  handle->Option_Descriptor[OPT_XDPI].size  = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_XDPI].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_xdpi);
  handle->Option_Descriptor[OPT_XDPI].constraint_type = SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_YDPI].name  = SANE_NAME_SCAN_Y_RESOLUTION;
  handle->Option_Descriptor[OPT_YDPI].title = SANE_TITLE_SCAN_Y_RESOLUTION;
  handle->Option_Descriptor[OPT_YDPI].desc  = SANE_DESC_SCAN_Y_RESOLUTION;
  handle->Option_Descriptor[OPT_YDPI].type  = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_YDPI].unit  = SANE_UNIT_DPI;
  handle->Option_Descriptor[OPT_YDPI].size  = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_YDPI].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_ydpi);
  handle->Option_Descriptor[OPT_YDPI].constraint_type = SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_ENHANCEMENT_GROUP].name  = "";
  handle->Option_Descriptor[OPT_ENHANCEMENT_GROUP].title = "Enhancement";
  handle->Option_Descriptor[OPT_ENHANCEMENT_GROUP].desc  = "";
  handle->Option_Descriptor[OPT_ENHANCEMENT_GROUP].type  = SANE_TYPE_GROUP;
  handle->Option_Descriptor[OPT_ENHANCEMENT_GROUP].cap   = 0;
  handle->Option_Descriptor[OPT_ENHANCEMENT_GROUP].constraint_type = 
    SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_BRIGHT].name  = SANE_NAME_BRIGHTNESS;
  handle->Option_Descriptor[OPT_BRIGHT].title = SANE_TITLE_BRIGHTNESS;
  handle->Option_Descriptor[OPT_BRIGHT].desc  = SANE_DESC_BRIGHTNESS;
  handle->Option_Descriptor[OPT_BRIGHT].type  = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_BRIGHT].unit  = SANE_UNIT_PERCENT;
  handle->Option_Descriptor[OPT_BRIGHT].size  = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_BRIGHT].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_bright);
  if (capabilities.bright.min == capabilities.bright.max) {
    handle->Option_Descriptor[OPT_BRIGHT].cap =
    handle->Option_Descriptor[OPT_BRIGHT].cap | SANE_CAP_INACTIVE;
  }
  handle->Option_Descriptor[OPT_BRIGHT].constraint_type = 
    SANE_CONSTRAINT_RANGE;
  if (!(range = malloc (sizeof(SANE_Range))))
    {
      destruct_Option_Descriptor_list (handle);
      return SANE_STATUS_NO_MEM;
    }
  range->min = capabilities.bright.min;
  range->max = capabilities.bright.max;
  handle->Option_Descriptor[OPT_BRIGHT].constraint.range = range;

  handle->Option_Descriptor[OPT_CONTRAST].name  = SANE_NAME_CONTRAST;
  handle->Option_Descriptor[OPT_CONTRAST].title = SANE_TITLE_CONTRAST;
  handle->Option_Descriptor[OPT_CONTRAST].desc  = SANE_DESC_CONTRAST;
  handle->Option_Descriptor[OPT_CONTRAST].type  = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_CONTRAST].unit  = SANE_UNIT_PERCENT;
  handle->Option_Descriptor[OPT_CONTRAST].size  = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_CONTRAST].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_contrast);
  if (capabilities.contrast.min == capabilities.contrast.max) {
    handle->Option_Descriptor[OPT_CONTRAST].cap =
    handle->Option_Descriptor[OPT_CONTRAST].cap | SANE_CAP_INACTIVE;
  }
  handle->Option_Descriptor[OPT_CONTRAST].constraint_type = 
    SANE_CONSTRAINT_RANGE;
  if (!(range = malloc (sizeof(SANE_Range))))
    {
      destruct_Option_Descriptor_list (handle);
      return SANE_STATUS_NO_MEM;
    }
  range->min = capabilities.contrast.min;
  range->max = capabilities.contrast.max;
  handle->Option_Descriptor[OPT_CONTRAST].constraint.range = range;

  handle->Option_Descriptor[OPT_HUE].name  = SANE_NAME_HUE;
  handle->Option_Descriptor[OPT_HUE].title = SANE_TITLE_HUE;
  handle->Option_Descriptor[OPT_HUE].desc  = SANE_DESC_HUE;
  handle->Option_Descriptor[OPT_HUE].type  = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_HUE].unit  = SANE_UNIT_PERCENT;
  handle->Option_Descriptor[OPT_HUE].size  = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_HUE].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_hue);
  if (capabilities.hue.min == capabilities.hue.max) {
    handle->Option_Descriptor[OPT_HUE].cap =
    handle->Option_Descriptor[OPT_HUE].cap | SANE_CAP_INACTIVE;
  }
  handle->Option_Descriptor[OPT_HUE].constraint_type = SANE_CONSTRAINT_RANGE;
  if (!(range = malloc (sizeof(SANE_Range))))
    {
      destruct_Option_Descriptor_list (handle);
      return SANE_STATUS_NO_MEM;
    }
  range->min = capabilities.hue.min;
  range->max = capabilities.hue.max;
  handle->Option_Descriptor[OPT_HUE].constraint.range = range;

  handle->Option_Descriptor[OPT_SAT].name  = SANE_NAME_SATURATION;
  handle->Option_Descriptor[OPT_SAT].title = SANE_TITLE_SATURATION;
  handle->Option_Descriptor[OPT_SAT].desc  = SANE_DESC_SATURATION;
  handle->Option_Descriptor[OPT_SAT].type  = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_SAT].unit  = SANE_UNIT_PERCENT;
  handle->Option_Descriptor[OPT_SAT].size  = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_SAT].cap = 
    LHII_CAP_2_SANE_CAP(capabilities.o_sat);
  if (capabilities.sat.min == capabilities.sat.max) {
    handle->Option_Descriptor[OPT_SAT].cap =
    handle->Option_Descriptor[OPT_SAT].cap | SANE_CAP_INACTIVE;
  }
  handle->Option_Descriptor[OPT_SAT].constraint_type = SANE_CONSTRAINT_RANGE;
  if (!(range = malloc (sizeof(SANE_Range))))
    {
      destruct_Option_Descriptor_list (handle);
      return SANE_STATUS_NO_MEM;
    }
  range->min = capabilities.sat.min;
  range->max = capabilities.sat.max;
  handle->Option_Descriptor[OPT_SAT].constraint.range = range;

  handle->Option_Descriptor[OPT_GEOMETRY_GROUP].name  = "";
  handle->Option_Descriptor[OPT_GEOMETRY_GROUP].title = "Geometry";
  handle->Option_Descriptor[OPT_GEOMETRY_GROUP].desc  = "";
  handle->Option_Descriptor[OPT_GEOMETRY_GROUP].type  = SANE_TYPE_GROUP;
  handle->Option_Descriptor[OPT_GEOMETRY_GROUP].cap   = 0;
  handle->Option_Descriptor[OPT_GEOMETRY_GROUP].constraint_type = 
    SANE_CONSTRAINT_NONE;

  handle->Option_Descriptor[OPT_LENGTH].name = "length";
  handle->Option_Descriptor[OPT_LENGTH].title = "scanning length";
  handle->Option_Descriptor[OPT_LENGTH].desc = "scanning length in mm";
  handle->Option_Descriptor[OPT_LENGTH].type = SANE_TYPE_INT;
  handle->Option_Descriptor[OPT_LENGTH].unit = SANE_UNIT_MM;
  handle->Option_Descriptor[OPT_LENGTH].size = sizeof(SANE_Int);
  handle->Option_Descriptor[OPT_LENGTH].cap = 
    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
  /*handle->Option_Descriptor[OPT_LENGTH].constraint_type = 
    SANE_CONSTRAINT_NONE;*/
  handle->Option_Descriptor[OPT_LENGTH].constraint_type = 
    SANE_CONSTRAINT_RANGE;
  if (!(range = malloc (sizeof(SANE_Range))))
    {
      destruct_Option_Descriptor_list (handle);
      return SANE_STATUS_NO_MEM;
    }
  range->min = 0;
  range->max = 300;
  range->quant = 1;
  handle->Option_Descriptor[OPT_LENGTH].constraint.range = range;

  handle->Option_Descriptor[OPT_ADDITIONAL_GROUP].name  = "";
  handle->Option_Descriptor[OPT_ADDITIONAL_GROUP].title = 
    "Additional Options:";
  handle->Option_Descriptor[OPT_ADDITIONAL_GROUP].desc  = "";
  handle->Option_Descriptor[OPT_ADDITIONAL_GROUP].type  = SANE_TYPE_GROUP;
  handle->Option_Descriptor[OPT_ADDITIONAL_GROUP].cap   = 0;
  handle->Option_Descriptor[OPT_ADDITIONAL_GROUP].constraint_type = 
    SANE_CONSTRAINT_NONE;

  options = malloc (capabilities.av_options*sizeof(struct scanner_option));

  status = get_scanner_options(handle->fd,options);

  /* additional options */
  for (i=0;i<capabilities.av_options;i++)
    {
      handle->Option_Descriptor[NUM_REG_OPT + i].name = 
	strdup(options[i].name);
      handle->Option_Descriptor[NUM_REG_OPT + i].title = 
	strdup(options[i].name);
      handle->Option_Descriptor[NUM_REG_OPT + i].desc = "";
      handle->Option_Descriptor[NUM_REG_OPT + i].type = SANE_TYPE_BOOL;
      handle->Option_Descriptor[NUM_REG_OPT + i].unit = SANE_UNIT_NONE;
      handle->Option_Descriptor[NUM_REG_OPT + i].size = sizeof(SANE_Bool);
      handle->Option_Descriptor[NUM_REG_OPT + i].cap = options[i].options;
      handle->Option_Descriptor[NUM_REG_OPT + i].cap = SANE_CAP_SOFT_SELECT;
      handle->Option_Descriptor[NUM_REG_OPT + i].constraint_type = 
	SANE_CONSTRAINT_NONE;
    }
  free (options);
  return status;
}

/********************************************************************/
/* delete_device_list                                               */
/* delete a list of sane_devices                                    */
/********************************************************************/
void
delete_device_list (SANE_Handle h)
{
  struct lhii_devstruct *handle = h;
  struct lhii_device_list_element** device_list;
  struct lhii_device_list_element *tmp0;
  struct lhii_device_list_element *tmp1;

  device_list = &handle->lhii_device_list;
  tmp0 = *device_list;
  while (tmp0) {
    free ((void*)tmp0->sane_device.name);
    free ((void*)tmp0->sane_device.vendor);
    free ((void*)tmp0->sane_device.model);
    tmp1 = tmp0;
    tmp0 = tmp0->next_device;
    free (tmp1);
  }
  if (handle->sane_device_list) 
    free(handle->sane_device_list);
}

/********************************************************************/
/* add_to_device_list                                               */
/* add a new entry to a list of sane_devices                        */
/********************************************************************/
SANE_Status
add_to_device_list (const char* dev_name,SANE_Handle h)
{
  int fd;
  struct lhii_devstruct *handle = h;
  struct scanner_type type;
  struct lhii_device_list_element** device_list;
  struct lhii_device_list_element *tmp;
  SANE_Status status;

  DBG_INIT();
  DBG(1, "add_to_device_list: begin\n");

  device_list = &handle->lhii_device_list;
  if (!*device_list) {
    DBG(1, "add_to_device_list: allocating first element\n");
    *device_list = malloc(sizeof(struct lhii_device_list_element));
    if (!*device_list) 
      return SANE_STATUS_NO_MEM;
    tmp = *device_list;
    tmp->next_device = NULL;
  }
  else {
    tmp = *device_list;
    while (tmp->next_device != NULL)
      tmp = tmp->next_device;
    tmp->next_device = malloc(sizeof(struct lhii_device_list_element));
    if (!tmp->next_device) {
      delete_device_list (device_list);
      return SANE_STATUS_NO_MEM;
    }
    tmp = tmp->next_device;
    tmp->next_device = NULL;
  }

  status = open_scanner_device(dev_name,&fd);
  status = get_scanner_type(fd,&type);
  tmp->sane_device.name = strdup(dev_name);
  tmp->sane_device.vendor = strdup(type.manufacturer);
  tmp->sane_device.model  = strdup(type.model);
  tmp->sane_device.type   = "handheld scanner";
  status = close_scanner_device(fd);
  DBG(1, "add_to_device_list: end\n");
  return status;
}

/********************************************************************/
/********************************************************************/
/****** "Registered" SANE API Functions *****************************/
/********************************************************************/
/********************************************************************/


/********************************************************************/
/* sane_init()                                                      */
/********************************************************************/
SANE_Status
sane_init (SANE_Int * version_code,SANE_Auth_Callback authorize)
{
  char dev_name[PATH_MAX];
  SANE_Status status = SANE_STATUS_IO_ERROR;
  size_t len;
  FILE *fp;

  DBG_INIT();
  DBG(1, "sane_init: begin\n");
  authorize = NULL;
  /* return the SANE version we got compiled under */
  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);

  if (!(lhii_sane_handle = malloc (sizeof(struct lhii_devstruct))))
    return SANE_STATUS_NO_MEM;

  memset(lhii_sane_handle, 0, sizeof(struct lhii_devstruct));

  /* parse config file */
  fp = sanei_config_open (LHII_CONFIG_FILE);
  if (!fp) {
    /* default to /dev/scanner instead of insisting on config file */
    DBG(1, "sane_init:  missing config file '%s'\n", LHII_CONFIG_FILE);
    status = add_to_device_list ("/dev/scanner",lhii_sane_handle);
    return status;
  }
  while (fgets(dev_name, sizeof (dev_name), fp)) {
    status = SANE_STATUS_GOOD;
    DBG(23, "sane_init:  config- %s", dev_name);
    if (dev_name[0] == '#') continue;	/* ignore comments */
    len = strlen (dev_name);
    if (dev_name[len - 1] == '\n') dev_name[--len] = '\0';
    if (!len) continue;			/* ignore empty lines */
    status = add_to_device_list (dev_name,lhii_sane_handle);
    if (status != SANE_STATUS_GOOD) break;
  }
  fclose (fp);
  
  DBG(1, "sane_init: end\n");
  return status;
}

/********************************************************************/
/* sane_open                                                        */
/********************************************************************/
SANE_Status 
sane_open (SANE_String_Const name, SANE_Handle * h)
{
  SANE_Status status = SANE_STATUS_IO_ERROR;
  struct lhii_devstruct *handle;

  DBG_INIT();
  DBG(1, "sane_open: begin\n");

  *h = lhii_sane_handle;
  handle = *h;
  if (strcmp(name,"") == 0 && handle->lhii_device_list != NULL)
    handle->dev_name = (void*)handle->lhii_device_list->sane_device.name;
  else
    handle->dev_name = (void*)name;

  DBG(1, "sane_open: %s\n",handle->dev_name);

  status = open_scanner_device(handle->dev_name,&handle->fd);
  status = construct_Option_Descriptor_list (handle);

  DBG(1, "sane_open: end\n");

  return status;
}

/********************************************************************/
/* sane_close                                                       */
/********************************************************************/
void 
sane_close (SANE_Handle h)
{
  struct lhii_devstruct *handle = h;
  close_scanner_device(handle->fd);
  handle->num_lines = 0;
  handle->scan_length = 0;
  handle->bytes2read = 0;
  memset(&handle->mode, 0, sizeof(struct modeinfo_struct));
  handle->mode.options = -1;
}

/********************************************************************/
/* sane_exit                                                        */
/********************************************************************/
void 
sane_exit (void)
{
  DBG_INIT();
  DBG(1, "sane_exit: begin\n");

  destruct_Option_Descriptor_list (lhii_sane_handle);
  DBG(1, "sane_exit: begin\n");
  delete_device_list (lhii_sane_handle);
  DBG(1, "sane_exit: begin\n");
  if (lhii_sane_handle)
    free(lhii_sane_handle);

  DBG(1, "sane_exit: end\n");
}

/********************************************************************/
/* sane_get_option_descriptor                                       */
/********************************************************************/
const SANE_Option_Descriptor * 
sane_get_option_descriptor (SANE_Handle h, SANE_Int n)
{
  struct lhii_devstruct *handle = h;
  if (handle->opt_count >= n)
    {
      return &handle->Option_Descriptor[n];
    }
  else
    return NULL;
}

/********************************************************************/
/* sane_control_option                                              */
/********************************************************************/
SANE_Status 
sane_control_option (SANE_Handle h, SANE_Int n,SANE_Action a,
		     void *v,SANE_Int * i)
{
  struct lhii_devstruct *handle = h;
  struct modeinfo_struct *mode;
  SANE_Status status = SANE_STATUS_IO_ERROR;

  mode = &handle->mode;


  switch (a) {
  case SANE_ACTION_GET_VALUE:
    status = get_scanner_mode(handle->fd,mode);
    if (n < NUM_REG_OPT)
      {
	switch (n) {
	case OPT_NUM_OPTS:
	  *(SANE_Int*)v = handle->opt_count;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_MONO:
	  if (mode->sc_mode == SCM_MONO)
	    *(SANE_Bool*)v = SANE_TRUE;
	  else
	    *(SANE_Bool*)v = SANE_FALSE;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_GREY:
	  if (mode->sc_mode == SCM_GREY)
	    *(SANE_Bool*)v = SANE_TRUE;
	  else
	    *(SANE_Bool*)v = SANE_FALSE;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_TRUE:
	  if (mode->sc_mode == SCM_TRUE)
	    *(SANE_Bool*)v = SANE_TRUE;
	  else
	    *(SANE_Bool*)v = SANE_FALSE;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_XDPI:
	  *(SANE_Int*)v = mode->xdpi;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_YDPI:
	  *(SANE_Int*)v = mode->ydpi;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_BRIGHT:
	  *(SANE_Int*)v = mode->bright;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_CONTRAST:
	  *(SANE_Int*)v = mode->contrast;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_HUE:
	  *(SANE_Int*)v = mode->hue;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_SAT:
	  *(SANE_Int*)v = mode->sat;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_LENGTH: 
	  *(SANE_Int*)v = handle->scan_length;
	  status = SANE_STATUS_GOOD;
	  break;
	default:
	  status = SANE_STATUS_UNSUPPORTED;
	}
      }
    else
      {
	if (mode->options == (n - NUM_REG_OPT))
	  *(SANE_Bool*)v = SANE_TRUE;
	else
	  *(SANE_Bool*)v = SANE_FALSE;
	status = SANE_STATUS_GOOD;
      }
    break;
  case SANE_ACTION_SET_VALUE:
    if (n < NUM_REG_OPT)
      {
	switch (n) {
	case OPT_MONO:
	  if (*(SANE_Bool*)v == SANE_TRUE)
	    mode->sc_mode = SCM_MONO;
	  break;
	case OPT_GREY:
	  if (*(SANE_Bool*)v == SANE_TRUE)
	    mode->sc_mode = SCM_GREY;
	  break;
	case OPT_TRUE:
	  if (*(SANE_Bool*)v == SANE_TRUE)
	    mode->sc_mode = SCM_TRUE;
	  break;
	case OPT_XDPI:
	  mode->xdpi = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_YDPI:
	  mode->ydpi = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_BRIGHT:
	  if (*(SANE_Int*)v < 
	      handle->Option_Descriptor[OPT_BRIGHT].constraint.range->min ||
	      *(SANE_Int*)v > 
	      handle->Option_Descriptor[OPT_BRIGHT].constraint.range->max)
	    return SANE_STATUS_INVAL;
	  mode->bright = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_CONTRAST:
	  if (*(SANE_Int*)v < 
	      handle->Option_Descriptor[OPT_CONTRAST].constraint.range->min ||
	      *(SANE_Int*)v > 
	      handle->Option_Descriptor[OPT_CONTRAST].constraint.range->max)
	    return SANE_STATUS_INVAL;
	  mode->contrast = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_HUE:
	  if (*(SANE_Int*)v < 
	      handle->Option_Descriptor[OPT_HUE].constraint.range->min ||
	      *(SANE_Int*)v > 
	      handle->Option_Descriptor[OPT_HUE].constraint.range->max)
	    return SANE_STATUS_INVAL;
	  mode->hue = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_SAT:
	  if (*(SANE_Int*)v < 
	      handle->Option_Descriptor[OPT_SAT].constraint.range->min ||
	      *(SANE_Int*)v > 
	      handle->Option_Descriptor[OPT_SAT].constraint.range->max)
	    return SANE_STATUS_INVAL;
	  mode->sat = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	case OPT_LENGTH:
	  handle->scan_length = *(SANE_Int*)v;
	  status = SANE_STATUS_GOOD;
	  break;
	default:
	  status = SANE_STATUS_UNSUPPORTED;
	}
      }
    else
      {
	if (*(SANE_Bool*)v == SANE_TRUE)
	  mode->options = n - NUM_REG_OPT;
      }
    if (n != OPT_LENGTH)
      {
	*i = SANE_INFO_RELOAD_OPTIONS;
	status = set_scanner_mode(handle->fd,mode);
      }
    break;
  default:
    status = SANE_STATUS_UNSUPPORTED;
  }

  return status;
} 

/********************************************************************/
/* sane_get_parameters                                              */
/********************************************************************/
SANE_Status 
sane_get_parameters (SANE_Handle h,SANE_Parameters * p)
{
  SANE_Status status = SANE_STATUS_IO_ERROR;
  struct modeinfo_struct *mode;
  struct lhii_devstruct *handle = h;

  mode = &handle->mode;
  status = get_scanner_mode(handle->fd,mode);
  switch (mode->sc_mode) {
  case SCM_MONO:
    p->format = SANE_FRAME_GRAY;
    p->bytes_per_line = mode->bpl;
    p->pixels_per_line = 8*mode->bpl;
    p->depth = 1;
    status = SANE_STATUS_GOOD;
    break;
  case SCM_GREY:
    p->format = SANE_FRAME_GRAY;
    p->bytes_per_line = mode->bpl;
    p->pixels_per_line = mode->bpl;
    p->depth = 8;
    status = SANE_STATUS_GOOD;
    break;
  case SCM_TRUE:
    p->format = SANE_FRAME_RGB;
    p->bytes_per_line = mode->bpl;
    p->pixels_per_line = mode->bpl/3;
    p->depth = 8;
    status = SANE_STATUS_GOOD;
    break;
  default:
    status = SANE_STATUS_IO_ERROR;
  }
  p->last_frame = SANE_TRUE;
  p->lines = handle->num_lines;

  return status;
}

/********************************************************************/
/* sane_start                                                       */
/********************************************************************/
SANE_Status 
sane_start (SANE_Handle h)
{
  struct lhii_devstruct *handle = h; 
  struct modeinfo_struct *mode;
  SANE_Status status = SANE_STATUS_IO_ERROR;

  mode = &handle->mode;

  status = set_scanner_mode(handle->fd,mode);
  status = get_scanner_mode(handle->fd,mode);
  handle->num_lines = handle->scan_length*mode->ydpi/25.4;
  handle->bytes2read = handle->num_lines*mode->bpl;
  return status;
}

/********************************************************************/
/* sane_cancel                                                      */
/********************************************************************/
void 
sane_cancel (SANE_Handle h)
{
  struct lhii_devstruct *handle = h; 
  SANE_Status status = SANE_STATUS_IO_ERROR;

  while (SANE_STATUS_GOOD != status) {
    status = close_scanner_device(handle->fd);
  }
  status = SANE_STATUS_IO_ERROR;
  while (SANE_STATUS_GOOD != status) {
    status = open_scanner_device(handle->dev_name,&handle->fd);
  }
}

/********************************************************************/
/* sane_read                                                        */
/********************************************************************/
SANE_Status 
sane_read (SANE_Handle h, SANE_Byte * buf,SANE_Int maxlen, SANE_Int * len)
{
  struct lhii_devstruct *handle = h;
  SANE_Status status = SANE_STATUS_IO_ERROR;

  *len = 0;
  if (handle->bytes2read > 0)
    {
      if((*len = read(handle->fd,buf,maxlen))>=0)
	{
	  status = SANE_STATUS_GOOD;
	  handle->bytes2read = handle->bytes2read - (*len);
	}
      else
	{
	  status = SANE_STATUS_IO_ERROR;
	}
    }
  else
    {
      status = SANE_STATUS_EOF;
    }

  return status;
}

/********************************************************************/
/* sane_get_devices                                                 */
/********************************************************************/
SANE_Status 
sane_get_devices (const SANE_Device *** device_list,SANE_Bool local_only)
{
  int i=0,num_devices=0;
  struct lhii_device_list_element *tmp;
  const SANE_Device **sane_device_list = NULL;
  struct lhii_devstruct *handle = lhii_sane_handle;

  DBG_INIT();
  tmp = handle->lhii_device_list;

  DBG(10, "sane_get_devices: begin\n");

  while (tmp != NULL)
    {
      tmp = tmp->next_device;
      num_devices++;
    }
  if (num_devices == 0)
    return SANE_STATUS_UNSUPPORTED;

  DBG(10, "sane_get_devices: num_devices: %d\n",num_devices);
  
  sane_device_list = malloc((num_devices + 1) * sizeof(SANE_Device*));
  if (!sane_device_list) 
    return SANE_STATUS_NO_MEM;

  tmp = handle->lhii_device_list;
  while (tmp != NULL)
    {
      sane_device_list[i] = &tmp->sane_device;
      tmp = tmp->next_device;
      i++;
    }
  
  sane_device_list[num_devices] = NULL;
  *device_list = sane_device_list;
  handle->sane_device_list = sane_device_list;
  DBG(10, "sane_get_devices: end\n");
  return SANE_STATUS_GOOD;
}

/* functions, which are not realy implemented yet*/

SANE_Status 
sane_get_select_fd (SANE_Handle h, SANE_Int *fd)
{
  return SANE_STATUS_UNSUPPORTED;
}

SANE_Status 
sane_set_io_mode (SANE_Handle h, SANE_Bool m)
{
  return SANE_STATUS_UNSUPPORTED;
}











